import os
import asyncio
import contextlib
import re
import sys


class Pty:
    def __init__(self):
        self._loop = asyncio.get_running_loop()
        self._parent, self.child = os.openpty()

    def close(self):
        self.close_parent()
        self.close_child()

    def close_parent(self):
        if self._parent is not None:
            os.close(self._parent)
            self._parent = None

    def close_child(self):
        if self.child is not None:
            os.close(self.child)
            self.child = None

    async def read(self, size):
        fut = self._loop.create_future()
        def reader():
            try:
                data = os.read(self._parent, size)
            except Exception as e:
                fut.set_exception(e)
            else:
                fut.set_result(data)
            finally:
                self._loop.remove_reader(self._parent)
        self._loop.add_reader(self._parent, reader)
        return await fut

    async def write(self, data):
        fut = self._loop.create_future()
        def writer():
            try:
                written = os.write(self._parent, data)
            except Exception as e:
                fut.set_exception(e)
            else:
                fut.set_result(written)
            finally:
                self._loop.remove_writer(self._parent)

        self._loop.add_writer(self._parent, writer)
        return await fut


class TermReader:
    def __init__(self, pty):
        self._pty = pty
        self._buf = bytearray()

    async def expect(self, search):
        while True:
            m = re.search(search, self._buf)
            if m:
                print("Found {}. Eaten: {}".format(search, self._buf[0:m.endpos]))
                del self._buf[0:m.endpos]
                return True
            try:
                self._buf.extend(await asyncio.wait_for(self._pty.read(4096), timeout=2))
            except asyncio.TimeoutError:
                print("Did not find {}. Available: {}".format(search, self._buf))
                return False


class Shell:
    @staticmethod
    def prepare_tty():
        os.setsid()
        # Force being controlled by terminal
        tmp_fd = os.open(os.ttyname(0), os.O_RDWR)
        os.close(tmp_fd)

    @staticmethod
    async def create(pty, init):
        environ = os.environ.copy()
        p = await asyncio.create_subprocess_exec(
            '/bin/bash', '--init-file', init, '-i',
            stdout=pty.child, stdin=pty.child, stderr=pty.child,
            close_fds=True,
            preexec_fn=Shell.prepare_tty,
            env=environ)
        pty.close_child()
        return Shell(p)

    def __init__(self, process):
        self._process = process

    async def kill(self):
        self._process.terminate()
        await asyncio.sleep(1)
        self._process.kill()

    async def aclose(self):
        killer = asyncio.create_task(self.kill())
        try:
            await asyncio.wait_for(self._process.wait(), timeout=1.5)
        except asyncio.TimeoutError:
            pass
        killer.cancel()


async def run(init, executable):
    with contextlib.closing(Pty()) as pty:
        async with contextlib.aclosing(await Shell.create(pty, init)):
            reader = TermReader(pty)
            assert await reader.expect(rb'prompt[$] ')
            print(await pty.write('{} -b \t\t'.format(executable).encode('ascii')))
            assert await reader.expect(b'\a') # Bell
            assert await reader.expect(rb'counterrevolutionary  *electroencephalogram  *uncharacteristically')


asyncio.run(run(sys.argv[1], sys.argv[2]))



#!/usr/bin/python3

import os
import sys

import dbus.service
import dbus.mainloop.glib
from gi.repository import GLib

BUS_NAME = "org.freedesktop.portal.Desktop"
OBJECT_PATH = "/org/freedesktop/portal/desktop"

NETWORK_MONITOR_IFACE = "org.freedesktop.portal.NetworkMonitor"


class Portal(dbus.service.Object):
    def __init__(self, connection, object_path, config):
        super(Portal, self).__init__(connection, object_path)
        self._config = config

    @dbus.service.method(
        dbus_interface=NETWORK_MONITOR_IFACE,
        in_signature="",
        out_signature="b",
    )
    def GetAvailable(self):
        return True

    @dbus.service.method(
        dbus_interface=NETWORK_MONITOR_IFACE,
        in_signature="",
        out_signature="b",
    )
    def GetMetered(self):
        return False

    @dbus.service.method(
        dbus_interface=NETWORK_MONITOR_IFACE,
        in_signature="",
        out_signature="u",
    )
    def GetConnectivity(self):
        return dbus.UInt32(4) # Full Network

    @dbus.service.method(
        dbus_interface=NETWORK_MONITOR_IFACE,
        in_signature="",
        out_signature="a{sv}",
    )
    def GetStatus(self):
        return dict(
            available=self.GetAvailable(),
            metered=self.GetMetered(),
            connectivity=self.GetConnectivity(),
        )

    @dbus.service.method(
        dbus_interface=NETWORK_MONITOR_IFACE,
        in_signature="su",
        out_signature="b",
    )
    def CanReach(self, hostname, port):
        return True


def main(argv):
    dbus.mainloop.glib.DBusGMainLoop(set_as_default=True)
    main_loop = GLib.MainLoop()

    bus = dbus.SessionBus()
    # Make sure we quit when the bus shuts down
    bus.add_signal_receiver(
        main_loop.quit,
        signal_name="Disconnected",
        path="/org/freedesktop/DBus/Local",
        dbus_interface="org.freedesktop.DBus.Local",
    )

    portal = Portal(bus, OBJECT_PATH, None)

    # Allow other services to assume our bus name
    bus_name = dbus.service.BusName(
        BUS_NAME, bus, allow_replacement=True, replace_existing=True, do_not_queue=True
    )

    main_loop.run()


if __name__ == "__main__":
    sys.exit(main(sys.argv))



#!/usr/bin/python3

import argparse
import http.client
import sys
import socket

class UnixSocketHTTPConnection(http.client.HTTPConnection):
    def __init__(self, socket_path):
        super().__init__('localhost')
        self._socket_path = socket_path

    def connect(self):
        s = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
        s.connect(self._socket_path)
        self.sock = s


def main(argv):
    parser = argparse.ArgumentParser('Call the snapd REST API')
    parser.add_argument('--socket', default='/run/snapd.socket',
                        help='The socket path to connect to')
    parser.add_argument('--method', default='GET',
                        help='The HTTP method to use')
    parser.add_argument('path', metavar='PATH',
                        help='The HTTP path to request')
    parser.add_argument('body', metavar='BODY', default=None, nargs='?',
                        help='The HTTP request body')
    args = parser.parse_args(argv[1:])

    conn = UnixSocketHTTPConnection(args.socket)
    conn.request(args.method, args.path, args.body)

    response = conn.getresponse()
    body = response.read()
    print(body.decode('UTF-8'))
    return response.status >= 300

if __name__ == '__main__':
    sys.exit(main(sys.argv))



import os
import re
import sys
import yaml


def die(s):
    print(s, file=sys.stderr)
    sys.exit(1)


def equals(name, s1, s2):
    if s1 != s2:
        die("in %s expected %r, got %r" % (name, s2, s1))


def matches(name, s, r):
    if not re.search(r, s):
        die("in %s expected to match %s, got %r" % (name, r, s))


def check(name, d, *a):
    ka = set()
    for k, op, *args in a:
        if op == maybe:
            d[k] = d.get(k, "")
        if k not in d:
            die("in %s expected to have a key %r" % (name, k))
        op(name + "." + k, d[k], *args)
        ka.add(k)
    kd = set(d)
    if ka < kd:
        die("in %s: extra keys: %r" % (name, kd - ka))


def exists(name, d):
    pass


def maybe(name, d):
    pass


verNotesRx = re.compile(r"^\w\S*\s+-$")


def verRevNotesRx(s):
    return re.compile(r"^\w\S*\s+\(\d+\)\s+[1-9][0-9]*\w+\s+" + s + "$")


def verRelRevNotesRx(s):
    return re.compile(
        r"^\w\S*\s+\d{4}-\d{2}-\d{2}\s+\(\d+\)\s+[1-9][0-9]*\w+\s+" + s + "$"
    )


if os.environ.get("SNAPPY_USE_STAGING_STORE", "") == "1":
    snap_ids = {
        "test-snapd-tools": "02AHdOomTzby7gTaiLX3M3SGMmXDfLJp",
        "test-snapd-devmode": "FcHyKyMiQh71liP8P82SsyMXtZI5mvVj",
        "test-snapd-python-webserver": "uHjTANBWSXSiYzNOUXZNDnOSH3POSqWS",
    }
else:
    snap_ids = {
        "test-snapd-tools": "eFe8BTR5L5V9F7yHeMAPxkEr2NdUXMtw",
        "test-snapd-devmode": "821MII7GAzoRnPvTEb8R51Z1s9e0XmK5",
        "test-snapd-python-webserver": "Wcs8QL2iRQMjsPYQ4qz4V1uOlElZ1ZOb",
    }

res = list(yaml.safe_load_all(sys.stdin))

equals("number of entries", len(res), 7)

check(
    "basic",
    res[0],
    ("name", equals, "basic"),
    ("summary", equals, "Basic snap"),
    ("path", matches, r"^basic_[0-9.]+_all\.snap$"),
    ("version", matches, verNotesRx),
    ("license", equals, "unset"),
    ("description", equals, "A basic buildable snap\n"),
    ("build-date", exists),
)

check(
    "basic-desktop",
    res[1],
    ("name", equals, "basic-desktop"),
    ("path", matches, "snaps/basic-desktop/$"),  # note the trailing slash
    ("summary", equals, ""),
    ("version", matches, verNotesRx),
    ("description", equals, "A basic snap with desktop apps\n"),
    ("license", equals, "GPL-3.0"),
    ("commands", exists),
)

check(
    "test-snapd-tools",
    res[2],
    ("name", equals, "test-snapd-tools"),
    ("publisher", matches, r"(Canonical✓|canonical)"),
    ("store-url", equals, "https://snapcraft.io/test-snapd-tools"),
    ("contact", equals, "snaps@canonical.com"),
    ("summary", equals, "Tools for testing the snapd application"),
    ("description", equals, "A tool to test snapd\n"),
    ("commands", exists),
    ("tracking", equals, "latest/stable"),
    ("installed", matches, verRevNotesRx("-")),
    ("refresh-date", exists),
    (
        "channels",
        check,
        ("latest/stable", matches, verRelRevNotesRx("-")),
        ("latest/candidate", equals, "↑"),
        ("latest/beta", equals, "↑"),
        ("latest/edge", matches, verRelRevNotesRx("-")),
        ("2.0/stable", exists),
        ("2.0/candidate", exists),
        ("2.0/beta", exists),
        ("2.0/edge", exists),
    ),
    ("snap-id", equals, snap_ids["test-snapd-tools"]),
    (
        "license",
        matches,
        r"(unknown|unset)",
    ),  # TODO: update once snap.yaml contains the right license
)

check(
    "test-snapd-devmode",
    res[3],
    ("name", equals, "test-snapd-devmode"),
    ("publisher", matches, r"(Canonical✓|canonical)"),
    ("store-url", equals, "https://snapcraft.io/test-snapd-devmode"),
    ("contact", equals, "snaps@canonical.com"),
    ("summary", equals, "Basic snap with devmode confinement"),
    (
        "description",
        equals,
        "A basic buildable snap that asks for devmode confinement\n",
    ),
    ("tracking", equals, "latest/beta"),
    ("installed", matches, verRevNotesRx("devmode")),
    ("refresh-date", exists),
    (
        "channels",
        check,
        ("latest/stable", equals, "–"),
        ("latest/candidate", equals, "–"),
        ("latest/beta", matches, verRelRevNotesRx("devmode")),
        ("latest/edge", matches, verRelRevNotesRx("devmode")),
    ),
    ("snap-id", equals, snap_ids["test-snapd-devmode"]),
    (
        "license",
        matches,
        r"(unknown|unset)",
    ),  # TODO: update once snap.yaml contains the right license
)

check(
    "core",
    res[4],
    ("name", equals, "core"),
    ("type", equals, "core"),  # attenti al cane
    ("publisher", exists),
    ("store-url", exists),
    ("summary", exists),
    ("description", exists),
    # tracking not there for local snaps
    ("tracking", maybe),
    ("installed", exists),
    ("refresh-date", exists),
    ("channels", exists),
    # contacts is set on classic but not on Ubuntu Core where we
    # sideload "core"
    ("contact", maybe),
    ("snap-id", maybe),
    (
        "license",
        matches,
        r"(unknown|unset)",
    ),  # TODO: update once snap.yaml contains the right license
)

check(
    "error", res[5], ("warning", equals, 'no snap found for "/etc/passwd"'),
)

# not installed snaps have "contact" information
check(
    "test-snapd-python-webserver",
    res[6],
    ("name", equals, "test-snapd-python-webserver"),
    ("publisher", matches, r"(Canonical✓|canonical)"),
    ("store-url", equals, "https://snapcraft.io/test-snapd-python-webserver"),
    ("contact", equals, "snaps@canonical.com"),
    ("summary", exists),
    ("description", exists),
    ("channels", exists),
    ("snap-id", equals, snap_ids["test-snapd-python-webserver"]),
    ("license", equals, "Other Open Source"),
)




#!/usr/bin/env python3

import argparse
import datetime
import os
import re
from typing import NamedTuple

import debian.changelog


def parse_arguments():
    parser = argparse.ArgumentParser(description="automatic changelog writer for snapd")
    parser.add_argument("version", type=str, help="new snapd version")
    parser.add_argument("lpbug", type=str, help="new snapd major release LP bug")
    parser.add_argument(
        "changelog",
        type=argparse.FileType("r"),
        help="path to new changelog entry as generated by snappy-dch",
    )
    return parser.parse_args()


class Distro(NamedTuple):
    name: str  # name of the distro in the packaging directory
    debian_name: str  # debian distribution name
    version_suffix: str  # suffix to add to the version number in changelogs


debianish_distros = [
    Distro("ubuntu-14.04", "trusty", "~14.04"),
    Distro("ubuntu-16.04", "xenial", ""),
    Distro("debian-sid", "unstable", "-1"),
]


other_distros = [
    "opensuse",
    "fedora",
    "arch",
]


def rewrite_version_number_file(file_name, pattern, version, write):
    """rewrite_version_number_file reads a packaging file with a version number
    in it and updates the version number to the new one specified using the
    regex pattern.
    It returns the new file contents and optionally writes out the new contents
    back to the file as well.
    """

    with open(file_name, "r") as fh:
        file_contents = fh.read()

    # replace the pattern (which should have one capturing group in it for the
    # version specifier pattern) with the captured group + the new version such
    # that we can keep any whitespace, etc in between the version specified and
    # the version number
    new_contents, n = re.subn(
        pattern, r"\g<1>" + version, file_contents, flags=re.MULTILINE
    )

    # check that we only did one replacement in the file
    if n > 1:
        raise RuntimeError(
            f"too many version patterns ({n}) matched in packaging file {file_name}"
        )
    elif n < 1:
        raise RuntimeError(f"version pattern not matched in packaging file {file_name}")

    if write is True:
        with open(file_name, "w") as fh:
            fh.write(new_contents)

    return new_contents


def update_fedora_changelog(opts, snapd_packaging_dir, new_changelog_entry, maintainer):
    spec_file = os.path.join(snapd_packaging_dir, "fedora", "snapd.spec")

    # rewrite the snapd.spec file with the right version
    spec_file_content = rewrite_version_number_file(
        spec_file,
        r"^(Version:\s+).*$",
        opts.version,
        False,
    )

    # now we also need to add the changelog entry to the snapd.spec file
    # this is a bit tricky, since we want a different format for the
    # changelog in snapd.spec than we have for debian, but luckily it's
    # just trimming whitespace off the front of each line in the
    # changelog

    dedented_changelog_lines = []
    for line in new_changelog_entry.splitlines():
        # strip the first 3 characters which are space characters so
        # that we only have one single whitespace
        dedented_changelog_lines.append(line[3:] + "\n")

    date = datetime.datetime.now().strftime("%a %b %d %Y")

    date_and_maintainer_header = f"* {date} {maintainer[0]} <{maintainer[1]}>\n"
    changelog_header = f"- New upstream release {opts.version}\n"
    fedora_changelog_lines = [
        date_and_maintainer_header,
        changelog_header,
    ] + dedented_changelog_lines

    # find the start of the changelog section in the rewritten spec file bytes
    changelog_section = "\n%changelog\n"
    idx = spec_file_content.find(changelog_section)
    if idx < 0:
        raise RuntimeError(
            "'%changelog' line in fedora spec file not found (was a comment or whitespace added to that line?)"
        )
    # rewrite the spec file using the replaced bits up to the changelog section,
    # then insert our new changelog entry lines, then add the rest of the
    # replaced bits of the spec file
    with open(spec_file, "w") as fh:
        # write the spec file up to and including the changelog section
        fh.write(spec_file_content[: idx + len(changelog_section)])
        # insert our new changelog entry
        for ch_line in fedora_changelog_lines:
            fh.write(ch_line)
        fh.write("\n")
        # write the rest of the original spec file
        fh.write(spec_file_content[idx + len(changelog_section) :])


def update_opensuse_changelog(
    opts, snapd_packaging_dir, new_changelog_entry, maintainer
):
    spec_file = os.path.join(snapd_packaging_dir, "opensuse", "snapd.spec")
    changes_file = os.path.join(snapd_packaging_dir, "opensuse", "snapd.changes")

    rewrite_version_number_file(
        spec_file,
        r"^(Version:\s+).*$",
        opts.version,
        True,
    )

    # add a template changelog to the changes file
    date = datetime.datetime.now(tz=datetime.timezone.utc).strftime(
        "%a %b %d %H:%M:%S %Z %Y"
    )

    email = maintainer[1]
    templ = f"""-------------------------------------------------------------------
{date} - {email}

- Update to upstream release {opts.version}

"""

    # read the existing changes file and then write the new changelog entry at
    # the top and then write the rest of the file
    with open(changes_file, "r") as fh:
        current = fh.read()
    with open(changes_file, "w") as fh:
        fh.write(templ)
        fh.write(current)


def write_github_release_entry(opts, new_changelog_entry):
    with open(f"snapd-{opts.version}-github-release.md", "w") as fh:
        # write the prefix header
        fh.write(
            f"""New snapd release {opts.version}

See https://forum.snapcraft.io/t/the-snapd-roadmap/1973 for high-level overview.

"""
        )

        # write the rest of the actual changelog
        for line in new_changelog_entry.splitlines():
            # strip the first 4 characters which are space characters so
            # that there's no leading prefix
            fh.write(line.lstrip())
            fh.write("\n")


def main(opts):
    this_script = os.path.realpath(__file__)
    snapd_root_git_dir = os.path.dirname(os.path.dirname(this_script))
    snapd_packaging_dir = os.path.join(snapd_root_git_dir, "packaging")

    # read all the changelog entries, expected to be formatted by snappy-dch
    new_changelog_entry = opts.changelog.read()

    # verify that the changelog entry lines are all in the right format
    for line_number, line in enumerate(new_changelog_entry.splitlines(), start=1):
        # each line should start with either 4 spaces, a - and then another
        # space, or 6 spaces
        if not line.startswith("    - ") and not line.startswith("      "):
            raise RuntimeError(
                f"unexpected changelog line format in line {line_number}"
            )
        if len(line) > 72:
            raise RuntimeError(
                f"line {line_number} too long, should wrap properly to next line"
            )

    # read the name and email of the person running the script using i.e. dch
    # conventions
    maintainer = debian.changelog.get_maintainer()

    # first handle all of the debian packaging files
    for distro in debianish_distros:
        debian_packaging_changelog = os.path.join(
            snapd_packaging_dir, distro.name, "changelog"
        )
        with open(debian_packaging_changelog) as fh:
            ch = debian.changelog.Changelog(fh)

        # setup a new block
        ch.new_block(
            package="snapd",
            version=opts.version + distro.version_suffix,
            distributions=distro.debian_name,
            urgency="medium",
            author=f"{maintainer[0]} <{maintainer[1]}>",
            date=debian.changelog.format_date(),
        )

        # add the new changelog entry with our standard header
        # the spacing here is manually adjusted, the top of the comment is always
        # the same
        templ = f"\n  * New upstream release, LP: #{opts.lpbug}\n" + new_changelog_entry
        ch.add_change(templ)

        # write it out back to the changelog file
        with open(debian_packaging_changelog, "w") as fh:
            ch.write_to_open_file(fh)

    # now handle all of the non-debian packaging files
    for distro in other_distros:
        if distro == "arch":
            # for arch all we need to do is change the PKGBUILD "pkgver" key
            rewrite_version_number_file(
                os.path.join(snapd_packaging_dir, "arch", "PKGBUILD"),
                r"^(pkgver=).*$",
                opts.version,
                True,
            )
        elif distro == "fedora":
            update_fedora_changelog(
                opts, snapd_packaging_dir, new_changelog_entry, maintainer
            )

        elif distro == "opensuse":
            update_opensuse_changelog(
                opts, snapd_packaging_dir, new_changelog_entry, maintainer
            )

    write_github_release_entry(opts, new_changelog_entry)


if __name__ == "__main__":
    opts = parse_arguments()
    main(opts)



#!/usr/bin/env python3

import apt
import re

best_golang = None
for p in apt.Cache():
    if re.match(r"golang-([0-9.]+)$", p.name):
        if (
            best_golang is None
            or apt.apt_pkg.version_compare(
                best_golang.candidate.version, p.candidate.version
            )
            < 0
        ):
            best_golang = p

print(best_golang.name)



#!/usr/bin/python3

from gi.repository import GLib
import dbus.mainloop.glib
import dbus.service
import sys

BUS_NAME = "io.netplan.Netplan"
OBJECT_PATH = "/io/netplan/Netplan"
DOC_IFACE = "io.netplan.Netplan"


class NetplanApplyService(dbus.service.Object):
    def __init__(self, connection, object_path, logfile):
        super().__init__(connection, object_path)
        self._logfile = logfile

    @dbus.service.method(dbus_interface=DOC_IFACE, in_signature="", out_signature="b")
    def Apply(self):
        # log that we were called and always return True
        with open(self._logfile, "a+") as fp:
            fp.write("Apply called\n")
        return True

    @dbus.service.method(
        dbus_interface=DOC_IFACE, in_signature="", out_signature="a(sv)"
    )
    def Info(self):
        # log that we were called and always return a dbus struct (i.e. python
        # tuple) with Features in it
        with open(self._logfile, "a+") as fp:
            fp.write("Info called\n")
        return [("Features", ["dhcp-use-domains", "ipv6-mtu"])]


def main(argv):
    logfile = argv[1]
    dbus.mainloop.glib.DBusGMainLoop(set_as_default=True)
    main_loop = GLib.MainLoop()

    bus = dbus.SystemBus()
    # Make sure we quit when the bus shuts down
    bus.add_signal_receiver(
        main_loop.quit,
        signal_name="Disconnected",
        path="/org/freedesktop/DBus/Local",
        dbus_interface="org.freedesktop.DBus.Local",
    )

    NetplanApplyService(bus, OBJECT_PATH, logfile)

    # Allow other services to assume our bus name
    dbus.service.BusName(
        BUS_NAME, bus, allow_replacement=True, replace_existing=True, do_not_queue=True
    )

    main_loop.run()


if __name__ == "__main__":
    sys.exit(main(sys.argv))



#!/usr/bin/env python3
import dbus
import sys


def run(bus):
    obj = bus.get_object("com.dbustest.HelloWorld", "/com/dbustest/HelloWorld")
    print(obj.SayHello(dbus_interface="com.dbustest.HelloWorld"))


if __name__ == "__main__":
    if sys.argv[1] == "system":
        bus = dbus.SystemBus()
    else:
        bus = dbus.SessionBus()
    run(bus)
